/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity.core

import com.skydoves.sandwich.ApiResponse
import com.skydoves.sandwich.StatusCode.NoContent
import com.skydoves.sandwich.getOrThrow
import com.skydoves.sandwich.ktor.executeApiResponse
import com.skydoves.sandwich.ktor.statusCode
import com.skydoves.sandwich.onSuccess
import edu.ucsc.its.temerity.BuildConfig
import edu.ucsc.its.temerity.BuildConfig.BUILD_DATE
import edu.ucsc.its.temerity.BuildConfig.BUILD_TIMEZONE
import edu.ucsc.its.temerity.BuildConfig.BUILD_VARIANT
import edu.ucsc.its.temerity.BuildConfig.LIB_VERSION
import edu.ucsc.its.temerity.api.PlatformApi
import edu.ucsc.its.temerity.api.TemerityApi
import edu.ucsc.its.temerity.api.sortByCreationDate
import edu.ucsc.its.temerity.core.Temerity.Companion.DEFAULT_WEB_TIMEOUT
import edu.ucsc.its.temerity.di.LibModule.libModule
import edu.ucsc.its.temerity.di.platformModule
import edu.ucsc.its.temerity.extensions.coroutines.availableThreads
import edu.ucsc.its.temerity.extensions.coroutines.createJobScope
import edu.ucsc.its.temerity.extensions.time.applyAuditLogFormat
import edu.ucsc.its.temerity.extensions.time.applyScheduledSessionDateFormat
import edu.ucsc.its.temerity.model.AuditLogEntry
import edu.ucsc.its.temerity.model.AuditLogSortOrder
import edu.ucsc.its.temerity.model.AuditLogSortOrder.NEW_FIRST
import edu.ucsc.its.temerity.model.BuildVariant
import edu.ucsc.its.temerity.model.Course
import edu.ucsc.its.temerity.model.Device
import edu.ucsc.its.temerity.model.DeviceRecordingSession
import edu.ucsc.its.temerity.model.EventType
import edu.ucsc.its.temerity.model.FolderPermissions.Action
import edu.ucsc.its.temerity.model.Group
import edu.ucsc.its.temerity.model.NewUser
import edu.ucsc.its.temerity.model.User
import edu.ucsc.its.temerity.model.UserGroup
import edu.ucsc.its.temerity.model.UserRecordingSession
import edu.ucsc.its.temerity.model.UserUpdate
import io.github.reactivecircus.cache4k.Cache
import io.ktor.client.HttpClient
import io.ktor.client.HttpClientConfig
import io.ktor.client.engine.HttpClientEngine
import io.ktor.client.plugins.HttpTimeout
import io.ktor.client.plugins.compression.ContentEncoding
import io.ktor.client.plugins.compression.ContentEncodingConfig
import io.ktor.client.plugins.defaultRequest
import io.ktor.client.plugins.logging.LogLevel.ALL
import io.ktor.client.plugins.logging.Logging
import io.ktor.client.request.header
import io.ktor.client.statement.HttpResponse
import io.ktor.http.URLProtocol
import io.ktor.util.cio.toByteArray
import io.ktor.utils.io.ByteReadChannel
import kotlinx.coroutines.CoroutineDispatcher
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.async
import kotlinx.coroutines.awaitAll
import kotlinx.coroutines.delay
import kotlinx.coroutines.withContext
import kotlinx.datetime.DateTimeUnit
import kotlinx.datetime.LocalDate
import kotlinx.datetime.LocalDateTime
import kotlinx.datetime.TimeZone
import kotlinx.datetime.minus
import kotlinx.serialization.SerializationException
import kotlinx.serialization.json.Json
import org.koin.core.Koin
import org.koin.core.KoinApplication
import org.koin.core.component.KoinComponent
import org.koin.core.component.get
import org.koin.core.logger.Level
import org.koin.core.logger.Level.DEBUG
import org.koin.core.logger.Level.ERROR
import org.koin.core.logger.Level.INFO
import org.koin.core.logger.Level.NONE
import org.koin.core.logger.Level.WARNING
import org.koin.core.logger.Logger
import org.koin.core.logger.MESSAGE
import org.koin.core.parameter.parametersOf
import org.koin.core.qualifier.named
import org.koin.dsl.koinApplication
import kotlin.time.Duration
import kotlin.time.Duration.Companion.minutes
import kotlin.time.DurationUnit.MILLISECONDS
import co.touchlab.kermit.Logger as KermitLogger
import io.ktor.client.plugins.logging.Logger as KtorLogger

/*
 * Adapted from https://github.com/santimattius/kmp-networking/blob/main/README.md
 */
public class Temerity internal constructor(
  private val config: TemClientConfig,
) : TemerityApi,
  KoinComponent {
  /**
   * Constructor for Temerity
   * @param temApiToken The [String] API token to use for the Temerity client
   */
  @Suppress("Unused")
  internal constructor(temApiToken: String) : this(TemClientConfig.withToken(temApiToken))

  /**
   * The following are the static public properties and utility functions of the Temerity library.
   * They should also be provided by any re-implementations of the library as companion object fields as well,
   * to allow for easy access to the version and build information without creating an instance of the library.
   */
  public companion object {
    @JvmStatic
    public val DEFAULT_WEB_TIMEOUT: Duration = 2.minutes

    // TODO: Use this as cache4k expiration time
    @JvmStatic
    public val DEFAULT_CACHE_EXPIRATION: Duration = 15.minutes

    @JvmStatic
    internal val DEFAULT_MINIMUM_THREAD_COUNT: Int = 2

    @JvmStatic
    internal fun createLogger(
      tag: String? = null,
      config: TemClientConfig? = null,
      supportKtxNotebook: Boolean = false,
    ): KermitLogger = LoggerFactory.createLogger(
      tag = tag,
      config = config,
      supportKtxNotebook = supportKtxNotebook,
    )
  }

  /**
   * The version of the Temerity library.
   * This is updated with each release of the library and should not be modified.
   * @return The version of the Temerity library as a semantic versioning-formatted [String]
   */
  @Suppress("MemberVisibilityCanBePrivate")
  public override val version: String = LIB_VERSION

  /**
   * The build variant of the library.
   * Either [BuildVariant.DEBUG] or [BuildVariant.RELEASE].
   * @return The build variant of the library as a [BuildVariant] enum
   */
  @Suppress("MemberVisibilityCanBePrivate")
  public override val buildVariant: BuildVariant = BuildVariant.valueOf(BUILD_VARIANT)

  /**
   * The date & time when the Temerity library was built.
   * @return The date & time when the Temerity library was built as a [LocalDateTime] object
   */
  @Suppress("MemberVisibilityCanBePrivate")
  public override val buildDate: LocalDateTime = LocalDateTime.parse(BUILD_DATE)

  /**
   * The timezone where the library was built.
   * @return The timezone where the library was built as a [TimeZone] object
   */
  @Suppress("MemberVisibilityCanBePrivate")
  public override val buildTimeZone: TimeZone = TimeZone.of(BUILD_TIMEZONE)

  /**
   * This function initializes the Koin dependency injection framework,
   * which provides the library's [HttpClient] and [PlatformApi] instances
   * via factories defined as part of the [libModule].
   */
  private fun createKoinApp(config: TemClientConfig) = object : TemerityKoinContext() {
    override val logger = createCommonLogger(tag = BuildConfig.PACKAGE_NAME, config)
    override val koinApp = koinApplication {
      logger(object : Logger() {
        override fun display(level: Level, msg: MESSAGE) {
          when (level) {
            DEBUG -> logger.d(msg)
            INFO -> logger.i(msg)
            ERROR -> logger.e(msg)
            NONE -> logger.v(msg)
            WARNING -> logger.w(msg)
          }
        }
      })
      modules(
        libModule(),
        platformModule(),
      )
    }
    override val koin: Koin = koinApp.koin
  }

  internal abstract class TemerityKoinContext {
    abstract val koinApp: KoinApplication
    abstract val koin: Koin
    abstract val logger: KermitLogger
  }

  /**
   * Kotlinx-serialization [Json] encoder/decoder object used for serializing/deserializing JSON object responses
   */
  private val json: Json = Json {
    coerceInputValues = true
  }
  internal val koinContext = createKoinApp(config)
  override fun getKoin(): Koin = koinContext.koin

  private var libraryLogger: KermitLogger
  private var platformApi: PlatformApi
  private var cachedUserRoleList: Cache<Int, String>
  private var libraryCoroutineDispatcher: CoroutineDispatcher
  private var libraryCoroutineScope: CoroutineScope
  private var jsonProcessingDispatcher: CoroutineDispatcher
  private var webRequestDispatcher: CoroutineDispatcher
  private var fileProcessDispatcher: CoroutineDispatcher

  init {
    check(!config.serviceToken.isNullOrBlank()) {
      "Service API auth token must be provided. Set it using TemClientConfig.serviceToken(token)"
    }

    check(!config.serviceUrl.isNullOrBlank()) {
      "Service url must be provided. Set it using TemClientConfig.serviceUrl(url)"
    }

    libraryCoroutineDispatcher = when (config.optUseVirtualThreads) {
      // Case 1: User requests to enable virtual threads
      // Need to check if JDK supports virtual threads
      true -> {
        val jvmVersion = Version.parse(BuildConfig.JVM_VERSION)
        if (jvmVersion >= Version(21, 0, 0)) {
          TODO("Configure virtual thread per task executor, create dispatcher")
        } else {
          get<CoroutineDispatcher>(named("libraryCoroutineDispatcher")) {
            parametersOf(availableThreads(config.optThreadCount), "Temerity Library Dispatcher")
          }
        }
      }
      // Case 2: Virtual threads are identified as unsupported by null value as developer constant, or user requests to disable them: null | false
      else -> {
        get<CoroutineDispatcher>(named("libraryCoroutineDispatcher")) {
          parametersOf(availableThreads(config.optThreadCount), "Temerity Library Dispatcher")
        }
      }
    }
    libraryCoroutineScope = get<CoroutineScope>(named("libraryCoroutineScope")) {
      parametersOf(libraryCoroutineDispatcher)
    }

    jsonProcessingDispatcher = get<CoroutineDispatcher>(named("childDispatcher")) { parametersOf(libraryCoroutineDispatcher, 2, "TemerityJsonProcessingCoroutineDispatcher") }
    webRequestDispatcher = get<CoroutineDispatcher>(named("childDispatcher")) { parametersOf(libraryCoroutineDispatcher, 2, "TemerityWebRequestCoroutineDispatcher") }
    fileProcessDispatcher = get<CoroutineDispatcher>(named("childDispatcher")) { parametersOf(libraryCoroutineDispatcher, 1, "TemerityFileProcessCoroutineDispatcher") }

    libraryLogger = get<KermitLogger>(named("libraryLogger")) {
      parametersOf(BuildConfig.PACKAGE_NAME, config)
    }

    platformApi = get<PlatformApi>(named("ktorfitApi")) {
      parametersOf(config, webRequestDispatcher)
    }

    cachedUserRoleList = Cache.Builder<Int, String>()
      .expireAfterWrite(config.optCacheTimeoutDuration ?: 15.minutes)
      .build()
  }

  /**
   * Exception class used to break out of a loop when reading paginated API responses.
   */
  internal class BreakException(val page: Int = -1) : RuntimeException() {
    fun lastPage(): Int = page
  }

  // Inlining advantageous here since copied parameter (ApiResponse<String>) is small and code fragment is used constantly (to validate every platform API request)
  private inline fun <reified T> decodeResponseCatching(response: ApiResponse<String>, page: Int = 0): T {
    // Execute platform request, throwing any network exceptions
    // TODO: Handle some network exceptions here, such as timeouts (for retry) or unauthorized (for invalid token)
    response.onSuccess {
      when (this.statusCode) {
        // If the API returns a 204 No Content status code, we should throw a BreakException to stop the page iteration loop
        // As of Cashew release, this is only used in getGroups() to stop the loop when no more groups are returned
        // May be used in the future for other paginated endpoints (such as getAuditLogEntries)
        NoContent -> throw BreakException(page - 1)

        else -> Unit
      }
    }
    val apiResponsePayload = response.getOrThrow()

    if (apiResponsePayload == "[]") throw BreakException(page - 1)
    // Handle deserialization exceptions. Queue up decode:
    val deserializedJson = createJobScope(jsonProcessingDispatcher).runCatching {
      json.decodeFromString<T>(apiResponsePayload)
    }
    // Attempt, returning the deserialized object if successful
    return deserializedJson.getOrElse { exception: Throwable ->
      when (exception) {
        is SerializationException -> {
          // TODO: Implement platform API version checking
          get<KermitLogger>(named("libraryLogger")).d { "Returned JSON object: ${exception.message}" }
          val error = "Encountered error decoding response from platform API. You likely need to choose a Temerity release that supports the API version implemented by your instance. \nPlatform API response data : $apiResponsePayload"
          error.run {
            get<KermitLogger>(named("libraryLogger")).e { this }
            error(this)
          }
        }

        else -> throw exception
      }
    }
  }

  public override suspend fun getUsers(): List<User> = withContext(libraryCoroutineDispatcher) {
    val userRequest = platformApi.getUsers()
    val response = userRequest.executeApiResponse<String>()
    decodeResponseCatching(response)
  }

  public override suspend fun getUser(userId: Long): User = withContext(libraryCoroutineDispatcher) {
    val returnedUsers = getUsers()
    returnedUsers.first { it.userId == userId }
  }

  public override suspend fun createUser(primaryIdentifier: String, newUser: NewUser): HttpResponse = withContext(libraryCoroutineDispatcher) {
    val validationMessage = newUser.validate()
    if (validationMessage != "Valid") {
      error(validationMessage)
    } else {
      val serializedNewUser = createJobScope(jsonProcessingDispatcher).run {
        json.encodeToString(newUser)
      }
      val request = platformApi.createUser(primaryIdentifier, serializedNewUser)
      request.getOrThrow()
    }
  }

  public override suspend fun updateUser(userId: Long, userUpdate: UserUpdate): HttpResponse = withContext(libraryCoroutineDispatcher) {
    val serializedUpdatedUser = createJobScope(jsonProcessingDispatcher).run {
      json.encodeToString(userUpdate)
    }
    val request = platformApi.setUser(userId, serializedUpdatedUser)
    request.getOrThrow()
  }

  public override suspend fun deleteUser(userId: Long): HttpResponse = withContext(libraryCoroutineDispatcher) {
    val request = platformApi.deleteUser(userId)
    request.getOrThrow()
  }

  public override suspend fun refreshCachedUserRoles(): List<String> = withContext(libraryCoroutineDispatcher) {
    val usersRequest = platformApi.getUsers()
    val returnedUsersResponse = usersRequest.executeApiResponse<String>().getOrThrow()
    val returnedUserList = createJobScope(jsonProcessingDispatcher).run {
      json.decodeFromString<List<User>>(returnedUsersResponse)
    }
    val roleTypes = returnedUserList.map { it.userType }.distinct()
    cachedUserRoleList.invalidateAll()
    roleTypes.forEach {
      cachedUserRoleList.put(it.hashCode(), it)
    }
    roleTypes
  }

  public override suspend fun getCachedUserRoles(refresh: Boolean): List<String> = withContext(libraryCoroutineDispatcher) {
    if (refresh) {
      refreshCachedUserRoles()
    }
    val returnedUserRoles = cachedUserRoleList.asMap().values.toList()
    returnedUserRoles.ifEmpty {
      refreshCachedUserRoles()
    }
  }

  public override suspend fun getUserGroups(userId: Long): List<UserGroup> = withContext(libraryCoroutineDispatcher) {
    val request = platformApi.getUserGroups(userId)
    val response = request.executeApiResponse<String>()
    decodeResponseCatching(response)
  }

  public override suspend fun getUserGroupsOwned(userId: Long): List<UserGroup> = withContext(libraryCoroutineDispatcher) {
    val request = platformApi.getUserGroupsOwned(userId)
    val response = request.executeApiResponse<String>()
    decodeResponseCatching<List<UserGroup>>(response)
  }

  public override suspend fun createUserWithExternalId(primaryIdentifier: String, newUser: NewUser): HttpResponse {
    TODO("Not yet implemented")
  }

  public override suspend fun addFolderPermission(perms: Action): HttpResponse {
    TODO("Not yet implemented")
  }

  public override suspend fun removeFolderPermission(perms: Action) {
    TODO("Not yet implemented")
  }

  public override suspend fun getMediaMetadata(linkageValue: String) {
    TODO("Not yet implemented")
  }

  public override suspend fun getGroups(paginated: Boolean): List<Group> = withContext(libraryCoroutineDispatcher) {
    when (paginated) {
      true -> {
        val returnedGroups = ArrayList<Group>()
        try {
          for (x in (0..Int.MAX_VALUE)) {
            val thisPageRequest = platformApi.getGroupPage(x, true)
            val response = thisPageRequest.executeApiResponse<String>()
            val thisPageGroups: List<Group> = decodeResponseCatching(response, page = x)
            returnedGroups.addAll(thisPageGroups)
          }
        } catch (e: Exception) {
          when (e) {
            is BreakException -> {
              if (config.optDebugEnabled) {
                get<KermitLogger>(named("libraryLogger")).d("Caught BreakException() from decodeResponseCatching() notifying we're done reading groups from API: ${e.page} pages read")
              }
              return@withContext returnedGroups
            }

            else -> throw e
          }
        }
        returnedGroups
      }

      else -> {
        emptyList()
      }
    }
  }

  public override suspend fun getCourse(courseCode: String): Course = withContext(libraryCoroutineDispatcher) {
    decodeResponseCatching(platformApi.getCourse(courseCode).executeApiResponse<String>())
  }

  public override suspend fun getCourses(paginated: Boolean): List<Course> = withContext(libraryCoroutineDispatcher) {
    when (paginated) {
      true -> {
        val returnedCourses = ArrayList<Course>()
        try {
          for (x in (0..Int.MAX_VALUE)) {
            // TODO: Implement user setting for default audit log sort order
            val thisPage = platformApi.getCoursePage(x, true)
            val thisPageCourses: List<Course> = decodeResponseCatching(thisPage.executeApiResponse<String>(), page = x)
            returnedCourses.addAll(thisPageCourses)
            delay(750)
          }
        } catch (e: Exception) {
          when (e) {
            is BreakException -> {
              if (config.optDebugEnabled) {
                get<KermitLogger>(named("libraryLogger")).d("Caught BreakException() from decodeResponseCatching() notifying we're done reading groups from API: $e")
              }
              return@withContext returnedCourses
            }

            else -> throw e
          }
        }
        returnedCourses
      }

      false -> {
        decodeResponseCatching(platformApi.getCourses().executeApiResponse<String>())
      }
    }
  }

  // Get all audit log entries for a list of event types within window
  // Defaults to getting audit log entries for all event types
  // TODO: Read sort order from settings
  public override suspend fun getAuditLogEntries(
    startTime: LocalDate,
    endTime: LocalDate,
    eventTypeList: List<EventType>?,
    sortOrder: AuditLogSortOrder?,
  ): List<AuditLogEntry> = withContext(libraryCoroutineDispatcher) {
    // TODO: Error out if user inputs time window > 1 month, since the platform API doesn't support
    val returnedEntries = ArrayList<AuditLogEntry>()
    val jobScope = createJobScope(libraryCoroutineScope.coroutineContext)
    val requestJobs = with(eventTypeList ?: EventType.entries) {
      map { eventType ->
        jobScope.async {
          getAuditLogEntries(startTime, endTime, eventType, sortOrder)
        }
      }
    }

    val results = requestJobs.awaitAll()
    results.forEach { returnedEntries.addAll(it) }

    // Apply passed sortOrder, or default to NEW_FIRST TODO: Read from user settings
    returnedEntries.applyOrDefault(sortOrder ?: NEW_FIRST)
    returnedEntries
  }

  // Get audit log entries for a specific event type within window
  // TODO: Read sort order from settings
  public override suspend fun getAuditLogEntries(
    startTime: LocalDate,
    endTime: LocalDate,
    eventType: EventType,
    sortOrder: AuditLogSortOrder?,
    paginated: Boolean,
  ): List<AuditLogEntry> = withContext(libraryCoroutineDispatcher) {
    require(startTime <= endTime) {
      "End date must be same or later than start date"
    }
    when (paginated) {
      true -> {
        val returnedEntries = ArrayList<AuditLogEntry>()
        var windowStart = if (endTime.minus(1, DateTimeUnit.MONTH) < startTime) {
          startTime
        } else {
          endTime.minus(1, DateTimeUnit.MONTH)
        }
        var windowEnd = endTime
        while (windowStart < windowEnd) {
          try {
            for (x in (0..Int.MAX_VALUE)) {
              val endpointRequest = platformApi.getAuditLogEntries(
                startTime = windowStart.applyAuditLogFormat(),
                endTime = windowEnd.applyAuditLogFormat(),
                entryOffset = x * 75,
                eventType = eventType.value,
              )
              val response = createJobScope(webRequestDispatcher).run {
                endpointRequest.executeApiResponse<String>()
              }
              val logEventEntries: List<AuditLogEntry> =
                decodeResponseCatching(response, page = x)
              returnedEntries.addAll(logEventEntries)
            }
          } catch (e: Exception) {
            when (e) {
              is BreakException -> {
                if (config.optDebugEnabled) {
                  get<KermitLogger>(named("libraryLogger")).d("Caught BreakException() from decodeResponseCatching() notifying we're done reading audit log entries from API for specific window: ${e.page} pages read")
                }
                break
              }

              else -> throw e
            }
          }
          windowEnd = windowStart
          val backStep = windowStart.minus(1, DateTimeUnit.MONTH)
          windowStart = if (backStep < startTime) startTime else backStep
        }
        // Apply passed sortOrder, or default to NEW_FIRST TODO: Read from user settings
        returnedEntries.applyOrDefault(sortOrder ?: NEW_FIRST)
        returnedEntries
      }

      false -> {
        TODO("Implement non-paginated audit log entry retrieval")
      }
    }
  }

  private suspend fun ArrayList<AuditLogEntry>.applyOrDefault(sortOrder: AuditLogSortOrder?): ArrayList<AuditLogEntry> = withContext(libraryCoroutineDispatcher) {
    with(this@applyOrDefault) {
      when (sortOrder) {
        null -> {
          // Currently reads from prefs, falling back to BuildConfig constant. Can be overridden by passing a sort order as a parameter.
          // TODO: Implement a way to pass a clientDefaultSortOrder to Builder
          // TODO: Persist user-defined default sort order for subsequent runs
          apply { sortByCreationDate(NEW_FIRST) }
        }

        else -> apply { sortByCreationDate(sortOrder) }
      }
    }
  }

  public override suspend fun getDevices(): List<Device> = withContext(libraryCoroutineDispatcher) {
    val request = platformApi.getDevices()
    val response = request.executeApiResponse<String>()
    decodeResponseCatching(response)
  }

  public override suspend fun getDevice(deviceId: Long): Device = withContext(libraryCoroutineDispatcher) {
    val request = platformApi.getDeviceById(deviceId)
    val response = request.executeApiResponse<String>()
    decodeResponseCatching(response)
  }

  public override suspend fun getUserSessions(
    userId: String,
    startTime: LocalDate,
    endTime: LocalDate,
  ): List<UserRecordingSession> = withContext(libraryCoroutineDispatcher) {
    val request = platformApi.getUserSessions(
      userId,
      startTime.applyScheduledSessionDateFormat(),
      endTime.applyScheduledSessionDateFormat(),
    )
    val response = request.executeApiResponse<String>()
    decodeResponseCatching(response)
  }

  public override suspend fun getDeviceSchedule(
    deviceId: Long,
    startTime: LocalDate,
    endTime: LocalDate,
  ): List<DeviceRecordingSession> = withContext(libraryCoroutineDispatcher) {
    try {
      val request = platformApi.getDeviceSchedule(
        deviceId,
        startTime.applyScheduledSessionDateFormat(),
        endTime.applyScheduledSessionDateFormat(),
      )
      val response = request.executeApiResponse<String>()
      return@withContext decodeResponseCatching(response)
    } catch (e: Exception) {
      when (e) {
        is BreakException -> {
          if (config.optDebugEnabled) {
            get<KermitLogger>(named("libraryLogger")).d("Caught BreakException() from decodeResponseCatching() notifying we're done reading recording sessions from API: ${e.page} pages read")
          }
          return@withContext emptyList<DeviceRecordingSession>()
        }

        else -> throw e
      }
    }
  }

  public override suspend fun getStorageAnalyticsReport(groupId: Long): ByteArray = withContext(libraryCoroutineDispatcher) {
    val apiRequest = platformApi.getStorageAnalyticsReport(groupId)
    val responseBytes = apiRequest.executeApiResponse<ByteReadChannel>().getOrThrow()
    withContext(fileProcessDispatcher) {
      responseBytes.toByteArray(limit = 2000000000)
    } // Limit file downloads to 2 GB
  }
}

/**
 * Adapted from https://github.com/joreilly/FantasyPremierLeague/
 * Adapted from the tmdb-api project: https://github.com/MoviebaseApp/tmdb-kotlin/raw/refs/heads/main/tmdb-api/src/commonMain/kotlin/app/moviebase/tmdb/core/HttpClientFactory.kt
 * This function creates an HttpClient with the provided HttpClientEngine.
 * The HttpClientEngine is injected by the Koin platformModule(), which as its name suggests provides an HttpClientEngine implementation that is platform-dependent.
 * It configures the client to log all requests and responses if enableNetworkLogs is true. TODO: read this as a build setting
 * It also sets a default request header with the provided authToken.
 * @param httpClientEngine The HttpClientEngine to use for the HttpClient.
 * @param config A block specifying the following options:
 *  - Whether to log all requests and responses.
 *  - Service URL to make requests to.
 *  - Token to set as a default request header.
 * @return The created HttpClient.
 */
internal fun buildHttpClient(
  httpClientEngine: HttpClientEngine,
  config: TemClientConfig,
  logger: KermitLogger,
): HttpClient {
  val defaultConfig: HttpClientConfig<*>.() -> Unit = {
    // Can't use install(ContentNegotiation){ json() } here because YuJa's API returns a 406 error if the Accept header is set to application/json
    // This requires JSON transformations to be done via manually-called library functions

    // Pass the stored authToken as a header in every request
    defaultRequest {
      header("authToken", config.serviceToken)
      url {
        protocol = URLProtocol.HTTPS
      }
    }

    if (config.optDebugEnabled) {
      when (config.optHttpClientLoggingBlock) {
        null -> {
          install(Logging) {
            this.logger = object : KtorLogger {
              override fun log(message: String) {
                logger.d(message)
              }
            }
            level = ALL
            sanitizeHeader { headerKey ->
              headerKey == "authToken"
            }
          }
        }
        else -> {
          config.optHttpClientLoggingBlock?.let {
            Logging(it)
          }
        }
      }
    }

    expectSuccess = config.optExpectSuccess

    if (config.optUseWebTimeout) {
      val defaultWebTimeout = DEFAULT_WEB_TIMEOUT.toLong(MILLISECONDS)
      val configuredWebTimeout = config.optWebTimeoutDuration?.toLong(MILLISECONDS)
      install(HttpTimeout) {
        connectTimeoutMillis = configuredWebTimeout ?: defaultWebTimeout
        requestTimeoutMillis = configuredWebTimeout ?: defaultWebTimeout
        socketTimeoutMillis = configuredWebTimeout ?: defaultWebTimeout
      }
    }

    if (config.optEnableCompression) {
      install(ContentEncoding) {
        mode = ContentEncodingConfig.Mode.CompressRequest
        deflate(1.0F)
        gzip(1.0F)
      }
    }

    config.optHttpClientConfigBlock?.invoke(this)
  }

  return config.optHttpClientBuilder?.invoke()?.config(defaultConfig) ?: HttpClient(httpClientEngine, defaultConfig)
}

/**
 * Create a Temerity instance using Kotlin-DSL.
 */
@TemDsl
public fun Temerity(block: TemClientConfig.() -> Unit): Temerity {
  //  Adapted from the tmdb-api project: https://github.com/MoviebaseApp/tmdb-kotlin/raw/refs/heads/main/tmdb-api/src/commonMain/kotlin/app/moviebase/tmdb/Tmdb4.kt
  val config = TemClientConfig().apply(block)
  return Temerity(config)
}

/**
 * Create a Temerity Client Configuration instance using Kotlin-DSL.
 */
@TemDsl
public fun TemClientConfig(block: TemClientConfig.() -> Unit): TemClientConfig = TemClientConfig().apply(block)
