/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity.di

import com.skydoves.sandwich.ktorfit.ApiResponseConverterFactory
import de.jensklingenberg.ktorfit.ktorfit
import edu.ucsc.its.temerity.api.PlatformApi
import edu.ucsc.its.temerity.api.createPlatformApi
import edu.ucsc.its.temerity.core.TemClientConfig
import edu.ucsc.its.temerity.core.Temerity.Companion.createLogger
import edu.ucsc.its.temerity.core.buildHttpClient
import edu.ucsc.its.temerity.extensions.coroutines.createLibraryScope
import io.ktor.client.HttpClient
import io.ktor.client.engine.HttpClientEngine
import kotlinx.coroutines.CoroutineDispatcher
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import org.koin.core.parameter.parametersOf
import org.koin.core.qualifier.named
import org.koin.dsl.module
import co.touchlab.kermit.Logger as KermitLogger

internal object LibModule {
  /**
   * This function provides the Koin module for the Temerity library.
   * It includes the platform module and provides factories for [HttpClient] and [PlatformApi].
   */
  internal fun libModule() = module {
    single<KermitLogger>(named("libraryLogger")) { (tag: String, config: TemClientConfig) ->
      createLogger(tag, config)
    }
    single<CoroutineDispatcher>(named("libraryCoroutineDispatcher")) { (threadCount: Int, dispatcherName: String) ->
      Dispatchers.IO.limitedParallelism(threadCount, dispatcherName)
    }
    factory<CoroutineDispatcher>(named("childDispatcher")) { (parentDispatcher: CoroutineDispatcher, threadCount: Int, dispatcherName: String) ->
      parentDispatcher.limitedParallelism(threadCount, dispatcherName)
    }
    single<CoroutineScope>(named("libraryCoroutineScope")) { (dispatcher: CoroutineDispatcher) ->
      createLibraryScope(dispatcher)
    }
    factory<HttpClient>(named("httpClient")) { (dispatcher: CoroutineDispatcher, config: TemClientConfig, kermit: co.touchlab.kermit.Logger) ->
      buildHttpClient(
        httpClientEngine = get<HttpClientEngine>(named("httpClientEngine")) { parametersOf(dispatcher) },
        config = config,
        logger = kermit,
      )
    }
    factory<PlatformApi>(named("ktorfitApi")) { (config: TemClientConfig, webRequestDispatcher: CoroutineDispatcher) ->
      val logger: KermitLogger = get<KermitLogger>(named("libraryLogger"))
      val client: HttpClient =
        get<HttpClient>(named("httpClient")) { parametersOf(webRequestDispatcher, config, logger) }
      val ktorfit = ktorfit {
        config.serviceUrl?.let { baseUrl(it) }
        httpClient(client)
        converterFactories(ApiResponseConverterFactory.create())
      }
      ktorfit.createPlatformApi()
    }
  }
}
