/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity.model

import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable

/*
 * Classes representing models used by the users API
 * https://support.yuja.com/hc/en-us/articles/360049580714-YuJa-API#h_01H8YCQKHZB4Q5S9GC1P7WWGG8
 */
@Serializable
public data class User(
  @SerialName("login_id")
  val loginId: String = "",

  @SerialName("email_address")
  val emailAddress: String = "",

  @SerialName("user_type")
  val userType: String = "",

  @SerialName("user_id")
  val userId: Long = -1,

  val timezone: String = "",

  @SerialName("custom_id")
  val customId: String = "",

  @SerialName("last_name")
  val lastName: String = "",

  @SerialName("phone_number")
  val phoneNumber: String = "",

  @SerialName("first_name")
  val firstName: String = "",
)

@Serializable
public data class NewUser(
  @SerialName("email_address")
  val emailAddress: String = "",

  @SerialName("first_name")
  val firstName: String = "",

  @SerialName("last_name")
  val lastName: String = "",

  @SerialName("login_id")
  val loginId: String = "",

  @SerialName("login_password")
  val loginPassword: String = "",

  @SerialName("custom_id")
  val customId: String = "",

  @SerialName("user_type")
  val userType: String = "",

  @SerialName("phone_number")
  val phoneNumber: String = "",

  @SerialName("primary_identifier")
  val primaryIdentifier: String = "",

  @SerialName("secondary_identifier")
  val secondaryIdentifier: String = "",
) {
  public fun validate(): String = when {
    emailAddress.isBlank() -> "Email address is required"
    firstName.isBlank() -> "First name is required"
    lastName.isBlank() -> "Last name is required"
    loginId.isBlank() -> "Login ID is required"
    loginPassword.isBlank() -> "Login password is required"
    userType.isBlank() -> "User type is required"
    else -> "Valid"
  }
}

/**
 * @param userType The type of user to create. The different types of users you can create are: "Student", "Instructor", "IT Manager"
 */
@Serializable
public data class UserUpdate(
  @SerialName("email_address")
  var emailAddress: String? = null,

  @SerialName("first_name")
  var firstName: String? = null,

  @SerialName("last_name")
  var lastName: String? = null,

  @SerialName("login_password")
  var loginPassword: String? = null,

  @SerialName("custom_id")
  var customId: String? = null,

  @SerialName("timezone")
  var timezone: String? = null,

  @SerialName("user_type")
  var userType: String? = null,

  @SerialName("phone_number")
  var phoneNumber: String? = null,
)

@Serializable
public data class UserGroup(
  @SerialName("publish_type")
  val publishType: String = "",

  @SerialName("is_active")
  val isActive: Int = -1,

  @SerialName("group_code")
  val groupCode: String = "",

  @SerialName("group_id")
  val groupId: Long = -1,

  @SerialName("group_name")
  val groupName: String = "",

  @SerialName("owner_id")
  val ownerId: Long = -1,

  @SerialName("group_security")
  val groupSecurity: String = "",

  @SerialName("group_term")
  val groupTerm: String = "",
)

/*
 * The different types of folder permission you can set are:
 *  - "Role"
 *  - "User"
 *  - "Group"
 */
@Serializable
public object FolderPermissions {
  //  public interface Action
  public sealed class Action {
    public class AddRole(
      public val folderId: Long,
      public val type: Type = Type.ROLE,
      public val name: UserRole,
      public val updateDescendants: Boolean = true,
      public val access: AccessLevel,
    ) : Action()
    public class AddUser(
      public val folderId: Long,
      public val type: Type = Type.USER,
      public val updateDescendants: Boolean = true,
      public val userName: String,
      public val access: AccessLevel,
    ) : Action()
    public class AddGroup(
      public val folderId: Long,
      public val type: Type = Type.GROUP,
      public val updateDescendants: Boolean = true,
      public val groupName: String,
      public val access: AccessLevel,
    ) : Action()
    public class RemoveRole(
      public val folderId: Long,
      public val type: Type = Type.ROLE,
      public val name: UserRole,
    ) : Action()
    public class RemoveUser(
      public val folderId: Long,
      public val type: Type = Type.USER,
      public val userName: String,
    ) : Action()
    public class RemoveGroup(
      public val folderId: Long,
      public val type: Type = Type.GROUP,
      public val groupName: String,
    ) : Action()
  }

  @Serializable
  public enum class Type(public val permissionGroup: String) {
    ROLE("Role"),
    USER("User"),
    GROUP("Group"),
    ;

    public companion object {
      public operator fun invoke(passedPermissionGroup: String): Type = entries.first { it.permissionGroup == passedPermissionGroup }
    }
  }

  @Serializable
  public enum class UserRole(public val roleType: String) {
    STUDENT("Student"),
    INSTRUCTOR("Instructor"),
    ;

    public companion object {
      public operator fun invoke(passedRoleType: String): UserRole = entries.first { it.roleType == passedRoleType }
    }
  }

  @Serializable
  public enum class AccessLevel(public val accessLevel: String) {
    FULL_ACCESS("Full Access"),
    EDIT_ACCESS("Edit Access"),
    EDIT_CAPTION_ACCESS("Edit Caption Access"),
    READ_ACCESS("Read Access"),
    ;

    public companion object {
      public operator fun invoke(passedAccessLevel: String): AccessLevel = entries.first { it.accessLevel == passedAccessLevel }
    }
  }
}

// @Serializable
// public data class AddRoleFolderPermissions(
//  val folderId: Long,
//  val type: FolderPermissions.Type = FolderPermissions.Type.ROLE,
//  val name: FolderPermissions.Roles,
//  val updateDescendants: Boolean = true,
//  val access: FolderPermissions.AccessLevel,
// ) : FolderPermissions.Action

// @Serializable
// public data class AddUserFolderPermissions(
//  val folderId: Long,
//  val type: FolderPermissions.Type = FolderPermissions.Type.USER,
//  val updateDescendants: Boolean = true,
//  val userName: String,
//  val access: FolderPermissions.AccessLevel,
// ) : FolderPermissions.Action

// @Serializable
// public data class AddGroupFolderPermissions(
//  val folderId: Long,
//  val type: FolderPermissions.Type = FolderPermissions.Type.GROUP,
//  val updateDescendants: Boolean = true,
//  val groupName: String,
//  val access: FolderPermissions.AccessLevel,
// ) : FolderPermissions.Action

// @Serializable
// public data class RemoveRoleFolderPermissions(
//  val folderId: Long,
//  val type: FolderPermissions.Type = FolderPermissions.Type.ROLE,
//  val name: FolderPermissions.Roles,
// ) : FolderPermissions.Action

// @Serializable
// public data class RemoveUserFolderPermissions(
//  val folderId: Long,
//  val type: FolderPermissions.Type = FolderPermissions.Type.USER,
//  val userName: String,
// ) : FolderPermissions.Action

// @Serializable
// public data class RemoveGroupFolderPermissions(
//  val folderId: Long,
//  val type: FolderPermissions.Type = FolderPermissions.Type.GROUP,
//  val groupName: String,
// ) : FolderPermissions.Action

@Serializable
public data class LinkageValue(
  val linkage: String,
)
